// Font Manager, version 2.3 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 2.0 -- Implemented object-oriented design; Separated font-
//                loading and displaying into separate routines; A font
//                registry now keeps all fonts in a central location,
//                which the text display routines tap into.
// Version 2.1 -- Added TextBox, for clipped, wrapped text in a
//                rectangle
// Version 2.2 -- Added g_Font(), g_FontColor(), and g_printf() for quick
//                memory-efficient font displaying
// Version 2.3 -- Added g_m*() routines for mono-spaced font displaying
// Version 2.5 -- Added multi-color font support (good for antialiasing)
//                Must use special palette made up of colors in groups of 16

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "vgl.h"
#include "font.h"

char *output;
FontManager FontMan;

int GlobalFontID=1, GlobalFontColor=15;
int GlobalFontX=0, GlobalFontY=0;
int GlobalFontW, GlobalFontH;
int GlobalTabStop=20;
char GlobalText[4096];

FontManager::FontManager()
{
  memset(Fonts, 0, sizeof(Fonts));
  for (int i=0; i<16; i++)
    FontType[i] = NO_FONT;
  memset(output, 0, sizeof(output));
  output = new char[2048];
  if (!output)
  {
    g_Done();
    printf("Font Manager: Error allocating output buffer.\n");
    exit(1);
  }
}

FontManager::~FontManager()
{
  int i;

  for (i=0; i<16; i++)
    KillFont(i);

  delete[] output;
}

void FontManager::LoadFont(char *f, int id, int type)
{
  if (!g_LoadBitmap(f, Fonts[id]))
  {
    g_Done();
    printf("Font Manager: Error loading font file %s.\n", f);
    exit(1);
  }
  FontType[id] = type;
}

void FontManager::KillFont(int id)
{
  g_KillBitmap(Fonts[id], 256);
  FontType[id] = NO_FONT;
}

int FontManager::SetFontSize(int size)
{
  int i, id, low, cur;

  cur = size;
  for (i=0; i<16; i++)
  {
    cur = abs(Fonts[i]['A'].y - size);
    if (cur<low)
    {
      low = cur;
      id = i;
    }
  }

  return id;
}

TextLine::TextLine(int id)
{
  Text = new char[2048];
  if (!Text)
  {
    g_Done();
    printf("Font Manager: Error allocating text buffer.\n");
    exit(1);
  }
  memset(Text, 0, sizeof(Text));
  SetFontID(id);
  SetSpacing(1);
  SetClear(0);
  SetTabs(20);
}

TextLine::~TextLine()
{
  delete[] Text;
}

int TextLine::StrLen(void)
{
  char *str = Text;

  int len = 0;
  while (*str)
  {
    switch (*str)
    {
      case TABSTOP :
        str++;
        break;
      case COLORCH :
        str++;
        break;
      default :
        if (*str != TABSTOP && *str != COLORCH) len += FontMan.CharWidth(FontID, *str)+CharSpacing;
        break;
    }
    str++;
  }

  return len;
}

void TextLine::SetText(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(Text, fmt, arglist);
  va_end(arglist);
}

void TextLine::Print(int x, int y, int c)
{
  char *str = Text;
  int def = c;

  if (BackClear) g_fRect(x, y, x+StrLen(), y+FontMan.FontHeight(FontID), BackClear);
  while (*str)
  {
    switch (*str)
    {
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          x = ((x/TabStop)*TabStop)+TabStop; else
          x = *str;
        break;
      case COLORCH :
        str++;
        if (*str==COLORCH) c = def; else
        {
          switch (FontMan.FontType[FontID])
          {
            case MONO_COLOR :
              c = *str;
              break;
            case MULTI_COLOR :
              c = *str * 16;
              break;
          }
        }
        break;
      default :
        g_ctmBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
        x += FontMan.CharWidth(FontID, *str)+CharSpacing;
    }
    str++;
  }
}

void TextLine::cenPrint(int y, int c)
{
  Print((g_ResX() - StrLen()) >> 1, y, c);
}

void TextLine::btwPrint(int x1, int x2, int y, int c)
{
  Print((((x2-x1+1) - StrLen()) >> 1)+x1, y, c);
}

void TextLine::sPrint(int x, int y, int c)
{
  char *str = Text;
  int def = c;

  if (BackClear) g_fRect(x, y, x+StrLen(), y+FontMan.FontHeight(FontID), BackClear);
  while (*str)
  {
    switch (*str)
    {
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          x = ((x/TabStop)*TabStop)+TabStop; else
          x = *str;
        break;
      case COLORCH :
        str++;
        if (*str==COLORCH) c = def; else
        {
          switch (FontMan.FontType[FontID])
          {
            case MONO_COLOR :
              c = *str;
              break;
            case MULTI_COLOR :
              c = *str * 16;
              break;
          }
        }
        break;
      default :
        g_ctmBitmap(x+1, y+1, 0, &FontMan.Fonts[FontID][*str]);
        switch (FontMan.FontType[FontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(x, y, &FontMan.Fonts[FontID][*str]);
            break;
        }
        x += FontMan.CharWidth(FontID, *str)+CharSpacing;
    }
    str++;
  }
}

void TextLine::scenPrint(int y, int c)
{
  sPrint((g_ResX() - StrLen()) >> 1, y, c);
}

void TextLine::sbtwPrint(int x1, int x2, int y, int c)
{
  sPrint((((x2-x1+1) - StrLen()) >> 1)+x1, y, c);
}

TextBox::TextBox(int x1, int y1, int x2, int y2, int id)
{
  Text = new char[64*1024];
  if (!Text)
  {
    g_Done();
    printf("Font Manager: Error allocating text buffer.\n");
    exit(1);
  }
  memset(Text, 0, sizeof(Text));
  SetFontID(id);
  SetSpacing(1);
  g_GetClip(OCX1, OCY1, OCX2, OCY2);
  SetClip(x1, y1, x2, y2);
  SetTabs(20);
}

TextBox::TextBox()
{
  Text = new char[64*1024];
  if (!Text)
  {
    g_Done();
    printf("Font Manager: Error allocating text buffer.\n");
    exit(1);
  }
  memset(Text, 0, sizeof(Text));
  SetFontID(0);
  SetSpacing(1);
  g_GetClip(OCX1, OCY1, OCX2, OCY2);
  SetClip(0, 0, 319, 199);
  SetTabs(20);
}

TextBox::~TextBox()
{
  delete[] Text;
}

void TextBox::SetClip(int x1, int y1, int x2, int y2)
{
  CX1   = x1;
  CY1   = y1;
  CX2   = x2;
  CY2   = y2;
  Width = x2-x1;
}

void TextBox::SetText(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(Text, fmt, arglist);
  va_end(arglist);

  // Wrap text

  char *str = Text;
  int CurWidth = 0;
  int stop = 0;
  
  while (*str)
  {
    switch (*str)
    {
      case '\n' :
        CurWidth = 0;
        break;
      case TABSTOP :
        str++;
        break;
      case COLORCH :
        str++;
        break;
      default :
        CurWidth += FontMan.CharWidth(FontID, *str)+CharSpacing;
        if (CurWidth > Width)
        {
          char *saved = str;
          // backtrack and stick a newline in first available space/dash
          stop = 0;
          while (*str-- && !stop)
          {
            if (*str==' ')
            {
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            } else
            if (*str=='-')
            {
              str++;
              memmove(str+1, str, strlen(Text));
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            } else
            if (*str=='\n' || str==Text)
            {
              str = saved;
              memmove(str+1, str, strlen(Text));
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            }
          }
        }
        break;
    }
    str++;
  }

  str = Text;
  Height = 0;
  while (*str)
  {
    if (*str=='\n') Height++;
    str++;
  }
}

int TextBox::LineHeight(char *text)
{
  char *str = text;
  int CurWidth = 0;
  int TextHeight = 0;
  int stop = 0;
  
  while (*str)
  {
    switch (*str)
    {
      case '\n' :
        CurWidth = 0;
        break;
      case TABSTOP :
        str++;
        break;
      case COLORCH :
        str++;
        break;
      default :
        CurWidth += FontMan.CharWidth(FontID, *str)+CharSpacing;
        if (CurWidth > Width)
        {
          char *saved = str;
          // backtrack and stick a newline in first available space
          stop = 0;
          while (*str-- && !stop)
          {
            if (*str==' ')
            {
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            } else
            if (*str=='-')
            {
              memmove(str+1, str, strlen(Text));
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            } else
            if (*str=='\n' || str==Text)
            {
              str = saved;
              memmove(str+1, str, strlen(Text));
              *str = '\n';
              stop = 1;
              CurWidth = 0;
            }
          }
        }
        break;
    }
    str++;
  }

  str = text;
  while (*str)
  {
    if (*str=='\n') TextHeight++;
    str++;
  }

  return TextHeight+1;
}

void TextBox::Print(int dy, int c)
{
  BITMAP *b;
  char *str = Text;
  char *s;
  int def=c;
  int x, y;

  g_SetClip(CX1, CY1, CX2, CY2);

  x = CX1;
  y = CY1-dy;

  while (*str)
  {
    switch (*str)
    {
      case '\n' :
        x = CX1;
        y += FontMan.FontHeight(FontID);
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          x = CX1+((x/TabStop)*TabStop)+TabStop; else
          x = CX1+*str;
        break;
      case COLORCH :
        str++;
        if (*str==COLORCH) c = def; else
        {
          switch (FontMan.FontType[FontID])
          {
            case MONO_COLOR :
              c = *str;
              break;
            case MULTI_COLOR :
              c = *str * 16;
              break;
          }
        }
        break;
      default :
        switch (FontMan.FontType[FontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(x, y, &FontMan.Fonts[FontID][*str]);
        }
        x += FontMan.CharWidth(FontID, *str)+CharSpacing;
        break;
    }
    str++;
  }
  g_SetClip(OCX1, OCY1, OCX2, OCY2);
}

void TextBox::sPrint(int dy, int c)
{
  BITMAP *b;
  char *str = Text;
  char *s;
  int def=c;
  int x, y;

  g_SetClip(CX1, CY1, CX2, CY2);

  x = CX1;
  y = CY1-dy;

  while (*str)
  {
    switch (*str)
    {
      case '\n' :
        x = CX1;
        y += FontMan.FontHeight(FontID);
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          x = CX1+((x/TabStop)*TabStop)+TabStop; else
          x = CX1+*str;
        break;
      case COLORCH :
        str++;
        if (*str==COLORCH) c = def; else
        {
          switch (FontMan.FontType[FontID])
          {
            case MONO_COLOR :
              c = *str;
              break;
            case MULTI_COLOR :
              c = *str * 16;
              break;
          }
        }
        break;
      default :
        g_ctmBitmap(x+1, y+1, 0, &FontMan.Fonts[FontID][*str]);
        switch (FontMan.FontType[FontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(x, y, c, &FontMan.Fonts[FontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(x, y, &FontMan.Fonts[FontID][*str]);
            break;
        }
        x += FontMan.CharWidth(FontID, *str)+CharSpacing;
        break;
    }
    str++;
  }
  g_SetClip(OCX1, OCY1, OCX2, OCY2);
}

void g_Font(int fontid)
{
  GlobalFontID = fontid;
  GlobalFontW = FontMan.FontWidth(fontid);
  GlobalFontH = FontMan.FontHeight(fontid);
}

int g_GetFont(void)
{
  return GlobalFontID;
}

void g_FontColor(int c)
{
  GlobalFontColor = c;
}

int g_GetFontColor(void)
{
  return GlobalFontColor;
}

int g_GetFontHeight(void)
{
  return FontMan.FontHeight(g_GetFont());
}

void g_gotoxy(int x, int y)
{
  GlobalFontX = x;
  GlobalFontY = y;
}

int g_FontX(void)
{
  return GlobalFontX;
}

int g_FontY(void)
{
  return GlobalFontY;
}

void g_SetTabs(int t)
{
  GlobalTabStop = t;
}

int g_GetTabs(void)
{
  return GlobalTabStop;
}

void g_printf(int x, int y, char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  GlobalFontX = x;
  GlobalFontY = y;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : //newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += FontMan.CharWidth(GlobalFontID, *str)+1;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_mprintf(int x, int y, char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  GlobalFontX = x;
  GlobalFontY = y;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : //newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }
        int bump = (GlobalFontW - FontMan.Fonts[GlobalFontID][*str].x)/2;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx+bump, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += GlobalFontW+1;
        break;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_sprintf(int x, int y, char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  GlobalFontX = x;
  GlobalFontY = y;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }

        g_ctmBitmap(newx+1, newy+1, 0, &FontMan.Fonts[GlobalFontID][*str]);
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += FontMan.CharWidth(GlobalFontID, *str)+1;
        break;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_msprintf(int x, int y, char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  GlobalFontX = x;
  GlobalFontY = y;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }

        int bump = (GlobalFontW - FontMan.Fonts[GlobalFontID][*str].x)/2;
        g_ctmBitmap(newx+bump+1, newy+1, 0, &FontMan.Fonts[GlobalFontID][*str]);
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx+bump, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += GlobalFontW+1;
        break;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_printf(char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }

        newx += FontMan.CharWidth(GlobalFontID, *str)+1;
        break;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_mprintf(char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }
        int bump = (GlobalFontW - FontMan.Fonts[GlobalFontID][*str].x)/2;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx+bump, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += GlobalFontW+1;
        break;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_sprintf(char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }

        g_ctmBitmap(newx+1, newy+1, 0, &FontMan.Fonts[GlobalFontID][*str]);
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += FontMan.CharWidth(GlobalFontID, *str)+1;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

void g_msprintf(char *fmt, ...)
{
  va_list arglist;
  int newx, newy;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  char *str = GlobalText;
  newx = GlobalFontX;
  newy = GlobalFontY;
  while (*str)
  {
    switch (*str)
    {
      case '\n' : // newline
        newx = 0;
        newy += GlobalFontH;
        break;
      case TABSTOP :
        str++;
        if (*str==TABSTOP)
          newx = ((newx/GlobalTabStop)*GlobalTabStop)+GlobalTabStop; else
          newx = *str;
        break;
      case COLORCH :
        str++;
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            GlobalFontColor = *str;
            break;
          case MULTI_COLOR :
            GlobalFontColor = *str * 16;
            break;
        }
        break;
      default :
        if (newx > g_ResX())
        {
          newx = 0;
          newy += GlobalFontH;
        }

        int bump = (GlobalFontW - FontMan.Fonts[GlobalFontID][*str].x)/2;
        g_ctmBitmap(newx+bump+1, newy+1, 0, &FontMan.Fonts[GlobalFontID][*str]);
        switch (FontMan.FontType[GlobalFontID])
        {
          case MONO_COLOR :
            g_ctmBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case MULTI_COLOR :
            g_ctiBitmap(newx+bump, newy, GlobalFontColor, &FontMan.Fonts[GlobalFontID][*str]);
            break;
          case FULL_COLOR :
            g_ctBitmap(newx+bump, newy, &FontMan.Fonts[GlobalFontID][*str]);
            break;
        }
        newx += GlobalFontW+1;
    }
    str++;
  }

  GlobalFontX = newx;
  GlobalFontY = newy;
}

int g_strlen(char *str)
{
  int len = 0;
  while (*str)
  {
    switch (*str)
    {
      case TABSTOP :
        str++;
        break;
      case COLORCH :
        str++;
        break;
      default :
        if (*str != TABSTOP && *str != COLORCH) len += FontMan.CharWidth(GlobalFontID, *str)+1;
        break;
    }
    str++;
  }

  return len;
}

int g_mstrlen(char *str)
{
  int len = 0;
  while (*str)
  {
    switch (*str)
    {
      case TABSTOP :
        str++;
        break;
      case COLORCH :
        str++;
        break;
      default :
        if (*str != TABSTOP && *str != COLORCH) len += GlobalFontW+1;
        break;
    }
    str++;
  }

  return len;
}

void g_cenprintf(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_strlen(GlobalText))/2;
  int newy = GlobalFontY;

  g_printf(newx, newy, GlobalText);
}

void g_mcenprintf(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_mstrlen(GlobalText))/2;
  int newy = GlobalFontY;

  g_mprintf(newx, newy, GlobalText);
}

void g_scenprintf(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_strlen(GlobalText))/2;
  int newy = GlobalFontY;

  g_sprintf(newx, newy, GlobalText);
}

void g_mscenprintf(char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_mstrlen(GlobalText))/2;
  int newy = GlobalFontY;

  g_msprintf(newx, newy, GlobalText);
}

void g_cenprintf(int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_strlen(GlobalText))/2;

  g_printf(newx, y, GlobalText);
}

void g_mcenprintf(int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_mstrlen(GlobalText))/2;

  g_mprintf(newx, y, GlobalText);
}

void g_scenprintf(int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_strlen(GlobalText))/2;

  g_sprintf(newx, y, GlobalText);
}

void g_mscenprintf(int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = (g_ResX() - g_mstrlen(GlobalText))/2;

  g_msprintf(newx, y, GlobalText);
}

void g_btwprintf(int x1, int x2, int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = ((x2-x1 - g_strlen(GlobalText))/2)+x1;

  g_printf(newx, y, GlobalText);
}

void g_mbtwprintf(int x1, int x2, int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = ((x2-x1 - g_mstrlen(GlobalText))/2)+x1;

  g_mprintf(newx, y, GlobalText);
}

void g_sbtwprintf(int x1, int x2, int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = ((x2-x1 - g_strlen(GlobalText))/2)+x1;

  g_sprintf(newx, y, GlobalText);
}

void g_msbtwprintf(int x1, int x2, int y, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(GlobalText, fmt, arglist);
  va_end(arglist);

  int newx = ((x2-x1 - g_mstrlen(GlobalText))/2)+x1;

  g_sprintf(newx, y, GlobalText);
}
