// The Z2 Game Kernel.

// The Kernel is responsible for keeping track of all objects, keeping track
// of the keyboard, and updating the game variables and game screen. It also
// interprets commands and sends messages between objects. In it's final form
// it will be completely game-independent, and can be used to do anything
// reasonable the author can imagine with the powerful scripting language.

// Watcom includes
#include <stdarg.h>
#include <stdio.h>
#include <time.h>

// HGL includes
#include "multikey.h"
#include "sprite.h"
#include "mouse.h"
#include "font.h"
#include "vgl.h"

// Z2 includes
#include "compiler.h"
#include "console.h"
#include "objects.h"
#include "editor.h"
#include "kernel.h"
#include "global.h"
#include "sound.h"
#include "user.h"

Z2TileMap RoomData;
pObject CamFocus;
pObject Player;
char buf[512], msg[512];
int msgfont, msgcolor, msgtime;
int barkfont, barkcolor, barktype;
int kDrawFloor, kDrawObject, kDrawCeiling;
int quitplaying, debug=0, hyper=0;
int CamX, CamY;

typedef struct {
  short send;                // Hash of label to jump to
  short line;                // Line in window
  short topline;             // # of lines
} tInter;

typedef struct {
  short send;                // Hash of label
  short line;                // Line in window
} tLabel;

void Bark(int t, pObject o, char *fmt, ...);
void Message(int k, int t, char *fmt, ...);
void DebugBark(char *fmt, ...);
void UpdateGame(void);

int Go(int d, pObject o);

void zFlip(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_S];

  for (int y=0; y<Tile_H; y++)
    memcpy(temp + (y*Tile_H), b.Data + ((Tile_H-1-y)*Tile_W), Tile_W);

  memcpy(b.Data, temp, Tile_S);
  delete[] temp;
}

void zMirror(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_S];

  for (int y=0; y<Tile_H; y++)
    for (int x=0; x<Tile_W; x++)
      temp[y*Tile_W+x] = b.Data[y*Tile_W+(Tile_H-1-x)];

  memcpy(b.Data, temp, Tile_S);
  delete[] temp;
}

void zRotateL(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_S];

  for (int y=0; y<Tile_H; y++)
    for (int x=0; x<Tile_W; x++)
      temp[y*Tile_W+x] = b.Data[x*Tile_W+(Tile_H-1-y)];

  memcpy(b.Data, temp, Tile_S);
  delete[] temp;
}

void zRotateR(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_S];

  for (int y=0; y<Tile_H; y++)
    for (int x=0; x<Tile_W; x++)
      temp[y*Tile_W+x] = b.Data[(Tile_H-1-x)*Tile_W+y];

  memcpy(b.Data, temp, Tile_S);
  delete[] temp;
}

void zShiftUp(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_W];

  memcpy(temp, b.Data, Tile_W);
  memmove(b.Data, b.Data+Tile_W, (Tile_H-1)*Tile_W);
  memcpy(b.Data + ((Tile_H-1)*Tile_W), temp, Tile_W);

  delete[] temp;
}

void zShiftDown(BITMAP &b)
{
  char *temp;

  temp = new char[Tile_W];

  memcpy(temp, b.Data + ((Tile_H-1)*Tile_W), Tile_W);
  memmove(b.Data+Tile_W, b.Data, (Tile_H-1)*Tile_W);
  memcpy(b.Data, temp, Tile_W);

  delete[] temp;
}

void zShiftLeft(BITMAP &b)
{
  char temp;

  for (int y=0; y<Tile_H; y++)
  {
    temp = b.Data[y*Tile_W];
    memmove(b.Data + (y*Tile_W), b.Data + (y*Tile_W+1), Tile_W-1);
    b.Data[y*Tile_W+(Tile_H-1)] = temp;
  }
}

void zShiftRight(BITMAP &b)
{
  char temp;

  for (int y=0; y<Tile_H; y++)
  {
    temp = b.Data[y*Tile_W+(Tile_H-1)];
    memmove(b.Data + (y*Tile_W+1), b.Data + (y*Tile_W), Tile_W-1);
    b.Data[y*Tile_W] = temp;
  }
}

int ResolveDir(char dir, short &x, short &y)
{
  // I sure hope the Watcom optimizer can reduce this... :)

  int reldir, pp=0;

  if (dir>=0x20 && dir<0x40) { pp = 1; dir -= 0x20; }
  if (dir>=0x40 && dir<0x60) { pp = 2; dir -= 0x40; }
  if (dir>=0x60 && dir<0x80) { pp = 3; dir -= 0x60; }
  if (dir>=0x80 && dir<0xA0) { pp = 4; dir -= 0x80; }

  // preprocess
  switch (dir)
  {
    case 0x0A : dir = (rand() % 4)+1; break;
    case 0x0B : dir = (rand() % 2)+1; break;  // random n/s
    case 0x0C : dir = (rand() % 2)+3; break;  // random e/w
    case 0x0D : break;  // random non-blocked
    case 0x0E : break;  // random blocked
    case 0x11 : break;  // flow
    case 0x12 : dir = GO_IDLE;
  }

  // process
  switch (dir)
  {
    case 0x01 : reldir = GO_NORTH; break;
    case 0x02 : reldir = GO_EAST;  break;
    case 0x03 : reldir = GO_SOUTH; break;
    case 0x04 : reldir = GO_WEST;  break;
  }

  // postprocess
  switch (pp)
  {
    case 1 : reldir = (reldir==GO_WEST) ? GO_NORTH : reldir+1; break;
    case 2 : reldir = (reldir==GO_NORTH) ? GO_WEST : reldir-1; break;
    case 3 :
      switch (reldir)
      {
        case GO_NORTH : reldir = GO_SOUTH; break;
        case GO_SOUTH : reldir = GO_NORTH; break;
        case GO_EAST  : reldir = GO_WEST;  break;
        case GO_WEST  : reldir = GO_EAST;  break;
      }
      break;
    case 4 :
      switch (reldir)
      {
        case GO_NORTH :
        case GO_SOUTH : dir = (rand() % 2)+3; break;  // random e/w
        case GO_EAST  :
        case GO_WEST  : dir = (rand() % 2)+1; break;  // random n/s
      }
      break;
  }

  // make final change
  switch (dir)
  {
    case GO_NORTH : y--; break;
    case GO_SOUTH : y++; break;
    case GO_EAST :  x++; break;
    case GO_WEST :  x--; break;
  }

  return reldir;
}

int x_barkfont(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num==1 || num==2) barkfont = num;
  return 5;
}

int x_barkcolor(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<256) barkcolor = num;
  return 5;
}

int x_bind(pObject o)
{
  int ps = strlen(o->code + o->codep + 2)+1;
  pObject Src = FindObj(o->code + o->codep + 2, SystemRoom->ObjHead);
  if (!Src) Src = FindObj(o->code + o->codep + 2, RoomCur->ObjHead);

  // note: if fail, display debug msg and return proper parameter size
  // if success, DO NOT pass proper parameter size (we are in new code)

  if (!Src)
  {
    DebugBark("#Bark: Object not found");
    return ps+2;
  } else
  {
    delete[] o->code;
    o->code = Src->code;
    o->codesize = Src->codesize;
    o->objBound = 1;
    o->codep = 0;
    return 0;
  }
}

int x_breakable(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  o->objBreakable = (opt != 0);
  return 3;
}

int x_ceiling(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  kDrawCeiling = opt;
  return 3;
}

int x_cycle(pObject o)
{
  memcpy(&o->cycle, o->code + o->codep + 2, sizeof(short));
//  o->cyclel = 0;
  return 4;
}

int x_cyclep(pObject o)
{
  memcpy(&Player->cycle, o->code + o->codep + 2, sizeof(short));
//  Player->cyclel = 0;
  return 4;
}

int x_dec(pObject o)
{
  int decby = 0;
  int ret = 0;
  pVar v = FindVar(o->code + o->codep + 3);
  o->code += 2 + strlen(o->code + o->codep + 2) + 1;

  if (o->code[o->codep]=='!')
  {
    decby = FindVarVal(o->code + o->codep + 1);
    ret += strlen(o->code + o->codep) + 1;
  } else
  {
    memcpy(&decby, o->code + o->codep, sizeof(short));
    ret += 2;
  }

  v->value -= decby;

  return ret+1;
}

int x_die(pObject o)
{
  KillObj(RoomCur, o);
  return 2;
}

int x_dieitem(pObject o)
{
  Player->x = o->x;
  Player->y = o->y;
  KillObj(RoomCur, o);
  return 2;
}

int x_end(pObject o)
{
  return 2;
}

int x_fadein(pObject o)
{
  zzt_ForceFade(Black, zztPal);
  return 2;
}

int x_fadeout(pObject o)
{
  zzt_ForceFade(zztPal, Black);
  return 2;
}

int x_floor(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  kDrawFloor = opt;
  return 5;
}

int x_frame(pObject o)
{
  if (o->code[o->codep + 2]==VarOp)
  {
    pVar v = FindVar(o->code + o->codep + 3);
    if (!v)
    {
      DebugBark("#Frame: Var '%s' not found", o->code + o->codep + 2);
      return 1;
    }
    o->frame = v->value;

    // 2 (for CMDx)
    // n (for var. name)
    // 1 (for var. name null)
    // 1 (for CMD null)
    return 2 + strlen(o->code + o->codep + 2) + 2;
  } else
  {
    memcpy(&o->frame, o->code + o->codep + 3, sizeof(short));
    return 5;
  }
}

int x_framep(pObject o)
{
  if (o->code[o->codep + 2]==VarOp)
  {
    pVar v = FindVar(o->code + o->codep + 3);
    if (!v)
    {
      DebugBark("#FrameP: Var '%s' not found", o->code + o->codep + 2);
      return 1;
    }
    if (Player) Player->frame = v->value;

    // 2 (for CMDx)
    // n (for var. name)
    // 1 (for var. name null)
    // 1 (for CMD null)
    return 2 + strlen(o->code + o->codep + 2) + 2;
  } else
  {
    if (Player) memcpy(&Player->frame, o->code + o->codep + 3, sizeof(short));
    return 5;
  }
}

int x_go(pObject o)
{
  char dir = o->code[o->codep + 2];
  short x = o->x;
  short y = o->y;
  int reldir = ResolveDir(dir, x, y);
  Go(reldir, o);
  return 4;
}

int x_goto(pObject o)
{
  return 1;
}

int x_idle(pObject o)
{
  return 2;
}

int x_if(pObject o)
{
  short lcomp, rcomp;
  pVar v;

  v = FindVar(o->code + o->codep + 3);
  if (!v)
  {
    DebugBark("#If: Var '%s' not found", o->code + o->codep + 2);
    return 1;
  }
  lcomp = v->value;

  int len = strlen(o->code + o->codep);
  int len2;

  if (o->code[o->codep + len + 3]==VarOp)
  {
    v = FindVar(o->code + o->codep + len + 4);
    if (!v)
    {
      DebugBark("#If: Var '%s' not found", o->code + o->codep + len + 3);
      return 1;
    }
    rcomp = v->value;
    len2 = strlen(o->code + o->codep + len + 4);
  } else
  {
    memcpy(&rcomp, o->code + o->codep + len + 2, sizeof(short));
    len2 = sizeof(short);
  }
  
  char op = o->code[o->codep + len + 1];
  len = len + len2 + 2;
  len2 = strlen(o->code + o->codep + len);
  switch (op)
  {
    case 0x01:
      if (lcomp==rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      break;
    case 0x02:
      if (lcomp!=rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      return 0;
    case 0x03:
      if (lcomp<=rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      return 0;
    case 0x04:
      if (lcomp>=rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      return 0;
    case 0x05:
      if (lcomp<rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      return 0;
    case 0x06:
      if (lcomp>rcomp)
        SendMessage(o, o->code + o->codep + len); else
        return len + len2 + 2;
      return 0;
  }

  return 0;
}

int x_inc(pObject o)
{
  int incby = 0;
  int ret = 0;
  pVar v = FindVar(o->code + o->codep + 3);
  if (!v)
  {
    DebugBark("#Inc: Var '%s' not found", o->code + o->codep + 2);
    return 1;
  }

  o->code += 2 + strlen(o->code + o->codep + 2) + 1;

  if (o->code[o->codep]==VarOp)
  {
    incby = FindVarVal(o->code + o->codep + 1);
    ret += strlen(o->code + o->codep) + 1;
  } else
  {
    memcpy(&incby, o->code + o->codep, sizeof(short));
    ret += 2;
  }

  v->value += incby;

  return ret+1;
}

int x_msgfont(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num==1 || num==2) msgfont = num;
  return 5;
}

int x_msgcolor(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<256) msgcolor = num;
  return 5;
}

int x_object(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  kDrawObject = opt;
  return 5;
}

int x_pushable(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  switch (opt)
  {
    case 0 :
      o->objPushNS = 0;
      o->objPushEW = 0;
      break;
    case 1 :
      o->objPushNS = 1;
      o->objPushEW = 1;
      break;
    case 2 :
      o->objPushNS = 1;
      o->objPushEW = 0;
      break;
    case 3 :
      o->objPushNS = 0;
      o->objPushEW = 1;
      break;
  }
  return 5;
}

int x_restart(pObject o)
{
  o->codep = 0;
  return 0;
}

int x_sendrel(pObject o)
{
  char dir = o->code[o->codep + 2];
  short x = o->x;
  short y = o->y;
  ResolveDir(dir, x, y);
  if (x>=0 && y>=0 && x<RoomCur->RoomW && y<RoomCur->RoomH)
  {
    pObject t = GetObjXY(x, y, RoomCur);
    SendMessage(t, o->code + o->codep + 3);
  }
  return 3 + strlen(o->code + o->codep + 3) + 2;
}

int x_set(pObject o)
{
  int set=0;
  int ret = 2 + strlen(o->code + o->codep + 2) + 1;
  int retval = ret;

  if (o->code[ret]==VarOp)
  {
    set = FindVarVal(o->code + ret + 1);
    retval += strlen(o->code + ret) + 1;
  } else
  {
    memcpy(&set, o->code + ret, sizeof(short));
    retval += 2;
  }

  AddVar(o->code + o->codep + 3, set, 1);

  return retval+1;
}

int x_shoot(pObject o)
{
  char dir = o->code[o->codep + 2];
  short x = o->x;
  short y = o->y;
  int reldir = ResolveDir(dir, x, y);

  switch (reldir)
  {
    case GO_NORTH: Spawn("BulletN", x, y); break;
    case GO_SOUTH: Spawn("BulletS", x, y); break;
    case GO_EAST:  Spawn("BulletE", x, y); break;
    case GO_WEST:  Spawn("BulletW", x, y); break;
  }

  return 4;
}

int x_solid(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  o->objSolid = (opt != 0);
  return 5;
}

int x_spawn(pObject o)
{
  short len = strlen(o->code + o->codep + 2);
  char  dir = o->code[o->codep + len + 3];

  short x = o->x;
  short y = o->y;
  ResolveDir(dir, x, y);

  Spawn(o->code + o->codep + 2, x, y);
  return 1+len+2;
}

int x_swimmer(pObject o)
{
  short opt;
  memcpy(&opt, o->code + o->codep + 2, sizeof(short));
  o->objSwimmer = (opt != 0);
  return 5;
}

int x_tileup(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zShiftUp(Tiles[num]);
  return 5;
}

int x_tiledown(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zShiftDown(Tiles[num]);
  return 5;
}

int x_tileleft(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zShiftLeft(Tiles[num]);
  return 5;
}

int x_tileright(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zShiftRight(Tiles[num]);
  return 5;
}

int x_tileflipv(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zFlip(Tiles[num]);
  return 5;
}

int x_tilefliph(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zMirror(Tiles[num]);
  return 5;
}

int x_tilerotatel(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zRotateL(Tiles[num]);
  return 5;
}

int x_tilerotater(pObject o)
{
  short num;
  memcpy(&num, o->code + o->codep + 2, sizeof(short));
  if (num>=0 && num<=TotTiles) zRotateR(Tiles[num]);
  return 5;
}

int x_walk(pObject o)
{
  return 0;
}

int x_openscroll(pObject o)
{
  int re_exec=0;
  ulong h;

  TextBox t(41, 41, 279, 160, 1);
  TextLine Title(1);
  int st, et;
  int LastPage;
  int dystart = -60;
  int quitscroll = 0;
  int dyend=0, olddy=0;
  int intnum=0, labnum=0;
  int curline=0, i=0, j=0;
  int px, py, TopBar, BotBar, DrawBar;

  // Initialize text (32k max)
  char *text = new char[32768];
  if (!text) return NO_MEM;
  memset(text, 0, sizeof(text));

  // Initialize label and interaction buttons
  tInter *Inter = new tInter[INTER_MAX];
  if (!Inter) return NO_MEM;
  tLabel *Lab = new tLabel[INTER_MAX];
  if (!Lab) return NO_MEM;

  // Save background
  g_CopyPage(CP, SwapPage);

  // open up a scroll window
  int FHeight = FontMan.FontHeight(1);
  int THeight = FHeight+2;

  Rect r;

  r.Set(34, 37-THeight, 252, THeight);
  Window TitleWin(r);

  r.Set(34, 39, 252, 123);
  Window TextWin(r);

  TitleWin.Draw();
  TitleWin.Save();
  TextWin.Draw();
  TextWin.Save();

  LastPage = g_GetCurPage();
  g_SetCurPage(SwapPage);

  // Feed text into buffer
  while (o->code[o->codep] != CommandOp &&
         o->code[o->codep] != BreakOp &&
         o->codep <= o->codesize)
  {
    switch (o->code[o->codep])
    {
      case TextOp :
        // Regular text is simply fed into the scroll buffer
        o->codep++;
        strcat(text, o->code + o->codep);
        strcat(text, "\n");
        curline += t.LineHeight(o->code + o->codep);
        o->codep += strlen(o->code + o->codep)+1;
        break;
      case CenTextOp :
        // Centered text is simply fed into the scroll buffer, with
        // a precalculated tab set.
        char tabcode[2];
        o->codep++;
        tabcode[0] = TABSTOP;
        tabcode[1] = (240-g_strlen(o->code + o->codep))/2;
        strncat(text, tabcode, 2);
        strcat(text, o->code + o->codep);
        strcat(text, "\n");
        curline += t.LineHeight(o->code + o->codep);
        o->codep += strlen(o->code + o->codep)+1;
        break;
      case InterOp :
        // Interaction buttons are kept track of, and fed into the
        // scroll buffer
        o->codep++;
        memcpy(&Inter[intnum].send, o->code + o->codep, sizeof(short));
        o->codep += 2;

        strcat(text, " ");
        strcat(text, o->code + o->codep);
        strcat(text, "\n");

        Inter[intnum].topline = (curline * FHeight)-60;
        curline += t.LineHeight(o->code + o->codep);
        o->codep += strlen(o->code + o->codep)+1;
        Inter[intnum++].line = (curline * FHeight)-60;
        break;
      case LabelOp :
        // Imbedded labels are kept track of
        o->codep++;
        Lab[labnum].send = Hash(o->code + o->codep);
        o->codep += strlen(o->code + o->codep)+1;
        Lab[labnum++].line = (curline * FHeight)-60;
        break;
      case ZapLabelOp :
      case CommentOp :
      case ObjNameOp :
        // Just skip past zapped labels, comments, object names...
        o->codep += strlen(o->code + o->codep)+1;
        break;
      default : o->codep++; break;
    }
  }
  // Set scroll buffer, and auto-wrap the text
  t.SetText(text);

  // Find ending boundary of wrapped text
  dyend = (t.GetHeight() * FHeight)-60;

  int accel=0;
  do {
    // Process keys
    if (mk_Down(mkHome)) dystart = -60;
    if (mk_Down(mkEnd))  dystart = dyend;

    if (mk_Shift() || OldScroll)
    {
      if (mk_Down(mkUp)) dystart -= FHeight/2;
      if (mk_Down(mkDown)) dystart += FHeight/2;
      if (mk_Down(mkPgUp)) dystart -= 120;
      if (mk_Down(mkPgDn)) dystart += 120;
    } else
    {
      if (mk_Down(mkUp))   if (accel>-30) accel -= 5;
      if (mk_Down(mkDown)) if (accel<30) accel += 5;
      if (mk_Down(mkPgUp)) accel = -36;
      if (mk_Down(mkPgDn)) accel = 36;
      dystart += accel/5;
      if (accel != 0)
      {
        if (accel<0) accel++;
        if (accel>0) accel--;
      }
    }

    // Keep it inbounds
    if (dystart<-60)   dystart = -60;
    if (dystart>dyend) dystart = dyend;

    // Only update the display if something has changed
    if (olddy != dystart)
    {
      Title.SetText("%s", o->name);

      // Check if cursor is above an interaction button
      DrawBar = 0;
      for (i=0; i<intnum; i++)
        if (dystart > Inter[i].topline && dystart <= Inter[i].line)
        {
//          Title.SetText("SendMessage (hash): %d", Inter[i].send);
          Title.SetText("Press Enter to select");
          TopBar  = Inter[i].topline;
          BotBar  = Inter[i].line;
          DrawBar = 1;
          break;
        }

      st = clock();

      // Save background
      TitleWin.Restore();
      TextWin.Restore();

      // Draw highlight bar
      if (DrawBar) g_ciRect(35, 100+(TopBar-dystart),
                            285, 100+(BotBar-dystart), WinDark);

      // fix annoying graphic glitch
      g_SetPixel(287, 37, colLGray);

      // Draw title and scroll text
      Title.scenPrint(37-THeight+1, 14);
      t.sPrint(dystart, 15);

      // Draw arrows on either side of scroll
      for (px=34, py=95; py<100; px++, py++)
        g_SetPixel(px, py, colWhite);
      for (px=38, py=101; py<106; px--, py++)
        g_SetPixel(px, py, colDGray);
      g_VLine(34, 96, 104, colLGray);
      g_VLine(35, 97, 103, colLGray);
      g_VLine(36, 98, 102, colLGray);
      g_VLine(37, 99, 101, colLGray);
      g_HLine(38, 39, 100, colLGray);

      for (px=286, py=95; py<100; px--, py++)
        g_SetPixel(px, py, colWhite);
      for (px=282, py=101; py<106; px++, py++)
        g_SetPixel(px, py, colDGray);
      g_VLine(286, 96, 104, colLGray);
      g_VLine(285, 97, 103, colLGray);
      g_VLine(284, 98, 102, colLGray);
      g_VLine(283, 99, 101, colLGray);
      g_HLine(281, 282, 100, colLGray);
//    g_HLine(40, 280, 100, 13);

      g_Update(SwapPage);
      do {
        et = clock();
      } while (et-st<1);

      olddy = dystart;
    }

    if (mk_Pressed(mkEsc))
    {
      re_exec = 0;
      quitscroll = 1;
      mk_Clear();
    }

    if (mk_Pressed(mkEnter))
    {
      // This unnecessarily complex chunk of code does several things.
      // 1. If the user pressed Enter, assume they want to quit.
      // 2. If the user pressed Enter on an Interaction, check for label.
      // 3. If label is imbedded, simply change starting position, cancel
      //    notice to quit, and continue the loop.
      // 4. If label is not imbedded, quit the loop, and issue a
      //    re-execution of the code.
      // 5. If the user pressed Enter on normal text, quit the loop, and
      //    do not issue a re-execution.
      re_exec = 0;
      quitscroll = 1;
      for (i=0; i<intnum; i++)
        if (dystart > Inter[i].line-FHeight && dystart <= Inter[i].line)
        {
          re_exec = RE_EXEC;
          for (j=0; j<labnum; j++)
          {
            if (Inter[i].send==Lab[j].send)
            {
              dystart = Lab[j].line+FHeight;
              re_exec = 0;
              quitscroll = 0;
            }
          }
          if (re_exec==RE_EXEC) SendMessage(o, Inter[i].send);
        }
      mk_Clear();
    }
  } while (!quitscroll);

  // Restore memory and set screen state as it was on entry
  g_SetCurPage(LastPage);
  TitleWin.Done();
  TextWin.Done();
  delete[] text;
  delete[] Inter;
  delete[] Lab;

  return re_exec;
}

int Execute(pObject o)
{
  int re_exec=0, add;
  if (!o) return BAD_OBJ;
  char ch, cmd;

  switch (o->code[o->codep])
  {
    case ObjNameOp :
    case LabelOp :
    case ZapLabelOp :
    case CommentOp :
      // Just skip past labels, comments, object names...
      o->codep += strlen(o->code + o->codep)+1;

      // Since a label or comment doesn't stop command processing, issue a
      // re-execution of code
      return re_exec = 1;
      break;
    case MoveOp :
      // move
      if (!o->queue) o->qsize = strlen(o->code + o->codep + 1);
    
      ch = o->code[o->codep + o->queue + 1];
      if (ch)
      {
        int r;
        switch (ch)
        {
          case 'n' : case 'N' : r = Go(GO_NORTH, o); break;
          case 's' : case 'S' : r = Go(GO_SOUTH, o); break;
          case 'e' : case 'E' : r = Go(GO_EAST, o); break;
          case 'w' : case 'W' : r = Go(GO_WEST, o); break;
          case 'i' : case 'I' : r = 0; break;
        }
        o->queue++;
        if (!r)
        {
          o->queue = 0;
          while (o->code[o->codep])
            o->codep++;
          o->codep++;
        }
      } else
      {
        o->queue = 0;
        while (o->code[o->codep])
          o->codep++;
        o->codep++;
      }

      return re_exec = 0;
      break;
    case PersgoOp :
      // persistent move
      return re_exec = 0;
      break;
    case CamOp :
      // camera
      return re_exec = 0;
      break;
    case BarkOp :
      // bark
      o->codep++;

      // message
      Bark(TM_DEF, o, o->code + o->codep);
      o->codep += strlen(o->code + o->codep)+1;

      return re_exec = 1;
      break;
    case PBarkOp :
      // player bark
      o->codep++;

      // message
      Bark(TM_DEF, Player, o->code + o->codep);
      o->codep += strlen(o->code + o->codep)+1;

      return re_exec = 1;
      break;
    case MessageOp :
      o->codep++;

      // message
      Message(BK_DEF, TM_DEF, o->code + o->codep);
      o->codep += strlen(o->code + o->codep)+1;

      return re_exec = 1;
      break;
    case MSMessageOp :
      o->codep++;

      // mono-spaced message
      Message(BK_MMSG, TM_DEF, o->code + o->codep);
      o->codep += strlen(o->code + o->codep)+1;

      return re_exec = 1;
      break;
    case CommandOp :
      cmd = o->code[o->codep+1];

      // process commands
      switch (cmd)
      {
        case CMD_BARKFONT     : add = x_barkfont(o); break;
        case CMD_BARKCOLOR    : add = x_barkcolor(o); break;
        case CMD_BIND         : add = x_bind(o); break;
        case CMD_BREAKABLE    : add = x_breakable(o); break;
        case CMD_CEILING      : add = x_ceiling(o); break;
        case CMD_CYCLE        : add = x_cycle(o); break;
        case CMD_CYCLEP       : add = x_cyclep(o); break;
        case CMD_DEC          : add = x_dec(o); break;
        case CMD_DIE          : add = x_die(o); break;
        case CMD_DIEITEM      : add = x_dieitem(o); break;
        case CMD_END          : add = x_end(o); break;
        case CMD_FADEIN       : add = x_fadein(o); break;
        case CMD_FADEOUT      : add = x_fadeout(o); break;
        case CMD_FLOOR        : add = x_floor(o); break;
        case CMD_FRAME        : add = x_frame(o); break;
        case CMD_FRAMEP       : add = x_framep(o); break;
        case CMD_GO           : add = x_go(o); break;
        case CMD_IDLE         : add = x_idle(o); break;
        case CMD_IF           : add = x_if(o); break;
        case CMD_INC          : add = x_inc(o); break;
        case CMD_MSGFONT      : add = x_msgfont(o); break;
        case CMD_MSGCOLOR     : add = x_msgcolor(o); break;
        case CMD_OBJECT       : add = x_object(o); break;
        case CMD_PUSHABLE     : add = x_pushable(o); break;
        case CMD_RESTART      : add = x_restart(o); break;
        case CMD_SENDREL      : add = x_sendrel(o); break;
        case CMD_SET          : add = x_set(o); break;
        case CMD_SHOOT        : add = x_shoot(o); break;
        case CMD_SOLID        : add = x_solid(o); break;
        case CMD_SPAWN        : add = x_spawn(o); break;
        case CMD_SWIMMER      : add = x_swimmer(o); break;
        case CMD_TILEUP       : add = x_tileup(o); break;
        case CMD_TILEDOWN     : add = x_tiledown(o); break;
        case CMD_TILELEFT     : add = x_tileleft(o); break;
        case CMD_TILERIGHT    : add = x_tileright(o); break;
        case CMD_TILEFLIPH    : add = x_tilefliph(o); break;
        case CMD_TILEFLIPV    : add = x_tileflipv(o); break;
        case CMD_TILEROTATEL  : add = x_tilerotatel(o); break;
        case CMD_TILEROTATER  : add = x_tilerotater(o); break;
      }
      switch (EOC[cmd])
      {
        case 0 : re_exec = 1; break;
        case 1 : re_exec = 0; break;
        case 3 : re_exec = 1; break;
        case 4 : re_exec = 0; break;
      }

      o->codep += add;
      return re_exec;
      break;
    case BreakOp :
      o->codep += 2;
      return re_exec = 0;
      break;
    case TextOp :
    case CenTextOp :
      return re_exec = x_openscroll(o);
      break;
  }

  return re_exec;
}

int UpdateObj(pObject o)
{
  if (!o) return BAD_OBJ;

  int re_exec;

  // We must execute at least once, so set re_exec to -1.
  re_exec = RE_EXEC;
  while (re_exec==RE_EXEC)
    re_exec = Execute(o);

  LastObj = o;
  return SUCCESS;
}

int Scan(pObject o, char *message)
{
  if (!o) return BAD_OBJ;
  int i=0;

  short hash = Hash(message);
  short comphash;

  while (i<o->codesize)
  {
    switch (o->code[i])
    {
      case LabelOp :
        i++;
        comphash = Hash(o->code + i);

        if (comphash==hash)
        {
          i += strlen(message)+1;
          o->codep = i;
          return SUCCESS;
        } else i += strlen(message)+1;
        break;
      case ObjNameOp :
      case TextOp :
      case CenTextOp :
      case ZapLabelOp :
      case CommentOp :
      case MoveOp :
      case PersgoOp :
      case CamOp :
      case BarkOp :
      case PBarkOp :
      case MessageOp :
      case MSMessageOp :
//      while (o->code[i)) i++;
//      i += strlen(o->code + i)+1;
        do { i++; } while (o->code[i]<0xF0);
        break;
      default :
//      i++;
        do { i++; } while (o->code[i]<0xF0);
        break;
    }
  }

  LastObj = o;
  return BAD_MSG;
}

int Scan(pObject o, short hash)
{
  if (!o) return BAD_OBJ;
  int i=0;
  short h;

  while (i<o->codesize)
  {
    switch (o->code[i])
    {
      case LabelOp :
        i++;
        h = Hash(o->code + i);
        if (h==hash)
        {
          i += strlen(o->code + i)+1;
          o->codep = i;
          return SUCCESS;
        } else i += strlen(o->code + i)+1;
        break;
      case CamOp :
      case MoveOp :
      case TextOp :
      case BarkOp :
      case PBarkOp :
      case PersgoOp :
      case ObjNameOp :
      case CenTextOp :
      case CommentOp :
      case MessageOp :
      case ZapLabelOp :
      case MSMessageOp :
//      while (o->code[i)) i++;
//      i += strlen(o->code + i)+1;
        do { i++; } while (o->code[i]<0xF0);
        break;
      default :
//      i++;
        do { i++; } while (o->code[i]<0xF0);
        break;
    }
  }

  LastObj = o;
  return BAD_MSG;
}

int SendPlayerMsg(char *message)
{
  if (Player) Scan(Player, message);
  return SUCCESS;
}

int SendGlobalMsg(char *message)
{
  pObject Cur = GlobalRoom->ObjHead;
  while (Cur)
  {
    Scan(Cur, message);
    Cur = Cur->next;
  }
  return SUCCESS;
}

int SendMessage(char *objname, char *message)
{
  // Check global objects
  if (strcmpi(objname, "player")==0)
  {
    return SendPlayerMsg(message);
  } else
  if (strcmpi(objname, "all")==0)
  {
    // every room object
    pObject Cur = RoomCur->ObjHead;

    while (Cur)
    {
      Scan(Cur, message);
      Cur = Cur->next;
    }
    return SUCCESS;
  } else
  {
    pObject o = FindObj(objname, RoomCur->ObjHead);
    if (!o) return BAD_OBJ;

    LastObj = o;
    return Scan(o, message);
  }
  return UNK_ERR;
}

int SendMessage(pObject o, char *message)
{
  if (!o) return BAD_OBJ;

  LastObj = o;
  return Scan(o, message);
}

int SendMessage(pObject o, short hash)
{
  if (!o) return BAD_OBJ;

  LastObj = o;
  return Scan(o, hash);
}

int SendMessage(char *roomname, char *objname, char *message)
{
  pRoom r = FindRoom(roomname);
  if (!r) return BAD_ROOM;

  pObject o = FindObj(objname, r->ObjHead);
  if (!o) return BAD_OBJ;

  LastObj = o;
  return Scan(o, message);
}

int SendSysMessage(char *objname, char *message)
{
  pObject o = FindObj(objname, SystemRoom->ObjHead);
  if (!o) return BAD_OBJ;

  LastObj = o;
  return Scan(o, message);
}

int SendHelpMessage(char *message)
{
  return Scan(Z2Help, message);
}

void UpdateCam(void)
{
  if (CamFocus)
  {
    if (CamX<CamFocus->x) CamX++;
    if (CamX>CamFocus->x) CamX--;
    if (CamY<CamFocus->y) CamY++;
    if (CamY>CamFocus->y) CamY--;
//    CamX = CamFocus->x;
//    CamY = CamFocus->y;
  }

  if (CamX<0) CamX = 0;
  if (CamX>RoomCur->RoomW) CamX = RoomCur->RoomW;
  if (CamY<0) CamY = 0;
  if (CamY>RoomCur->RoomH) CamY = RoomCur->RoomH;
}

void Update(void)
{
  // Update global objects
  pObject Cur = GlobalRoom->ObjHead;
  while (Cur)
  {
    Cur->cyclel++;
    if (Cur->cyclel==Cur->cycle)
    {
      Cur->cyclel = 0;
      UpdateObj(Cur);
    }
    Cur = Cur->next;
  }

  // Update room objects
  Cur = RoomCur->ObjHead;
  while (Cur)
  {
    Cur->cyclel++;
    if (Cur->cyclel==Cur->cycle)
    {
      Cur->cyclel = 0;
      UpdateObj(Cur);
    }
    Cur = Cur->next;
  }
}

void InitGame(void)
{
  CamX = CamY = 0;
  Player = NULL;
}

void UpdateGame(void)
{
  int AbsMapX, AbsMapY, i;
  pObject Cur;

  g_SetCurPage(1);
  g_SuperBlack();

  AbsMapX = (CamX*Tile_W)-160;
  AbsMapY = (CamY*Tile_H)-96;

  if (RoomCur->RoomW<320/Tile_W)
  {
    AbsMapX = ((RoomCur->RoomW*Tile_W)-320l)/2l;
  } else
  {
    if (AbsMapX<0) AbsMapX = 0;
    if (AbsMapX>(RoomCur->RoomW*Tile_W)-320) AbsMapX = (RoomCur->RoomW*Tile_W)-320;
  }

  if (RoomCur->RoomH<200/Tile_H)
  {
    AbsMapY = ((RoomCur->RoomH*Tile_H)-200l)/2l;
  } else
  {
    if (AbsMapY<0) AbsMapY = 0;
    if (AbsMapY>(RoomCur->RoomH*Tile_H)-192) AbsMapY = (RoomCur->RoomH*Tile_H)-192;
  }

  RoomData.SetPos(AbsMapX, AbsMapY);
  if (kDrawFloor) RoomData.DrawBase();
  if (kDrawObject)
  {
    RoomData.DrawFringe();

    // draw separate objects
    Cur = RoomCur->ObjHead;
    while (Cur)
    {
      if (!Cur->objInvisible)
      {
        int x = RoomData.Tile2ScrX(Cur->x);
        int y = RoomData.Tile2ScrY(Cur->y);
        if (x>-Tile_W && x<320 && y>-Tile_H && y<200)
        {
          if (x>0 && x<320-Tile_W && y>0 && y<200-Tile_H)
            g_tBitmap(x, y, &Tiles[Cur->frame]); else
            g_ctBitmap(x, y, &Tiles[Cur->frame]);
        } // don't draw way out of bounds
      }
      Cur = Cur->next;
    }
  }

  if (RoomCur->RoomCeiling && kDrawCeiling) RoomData.DrawRoof();

  // update barks, messages, etc...
  Cur = RoomCur->ObjHead;
  g_Font(2);
  while (Cur)
  {
    if (Cur->barktime)
    {
      g_FontColor(Cur->barktime % 16);
      int bx = RoomData.Tile2ScrX(Cur->x);
      int by = RoomData.Tile2ScrY(Cur->y);
      if (by<8)
        g_sprintf(bx+Tile_W+2, by, Cur->bark); else
        g_sprintf(bx, by-8, Cur->bark);
      Cur->barktime--;
    }
    Cur = Cur->next;
  }

  if (msgtime)
  {
    g_Font(1);
    g_FontColor(msgtime % 16);
    switch (barktype)
    {
      case 2 : // message
        g_scenprintf(190, msg);
        break;
      case 3 : // mono-spaced message
        g_mscenprintf(190, msg);
        break;
    }
    msgtime--;
  }

  g_HSync();
  g_Update();
}

void SetupWindowKernel(void)
{
  int AbsMapX, AbsMapY, i;
  pObject Cur;

  g_SetCurPage(SpriPage);
  g_SuperBlack();

  AbsMapX = (CamX*Tile_W)-160;
  AbsMapY = (CamY*Tile_H)-96;

  if (RoomCur->RoomW<(320/Tile_W) && RoomCur->RoomH<(200/Tile_H))
  {
    AbsMapX = ((RoomCur->RoomW*Tile_W)-320l)/2l;
    AbsMapY = ((RoomCur->RoomH*Tile_H)-200l)/2l;
  } else
  {
    if (AbsMapX<0) AbsMapX = 0;
    if (AbsMapY<0) AbsMapY = 0;
    if (AbsMapX>(RoomCur->RoomW*Tile_W)-320) AbsMapX = (RoomCur->RoomW*Tile_W)-320;
    if (AbsMapY>(RoomCur->RoomH*Tile_H)-192) AbsMapY = (RoomCur->RoomH*Tile_H)-192;
  }

  RoomData.SetPos(AbsMapX, AbsMapY);
  if (kDrawFloor) RoomData.DrawBase();
  if (kDrawObject)
  {
    RoomData.DrawFringe();

    // draw separate objects
    Cur = RoomCur->ObjHead;
    while (Cur)
    {
      int x = RoomData.Tile2ScrX(Cur->x);
      int y = RoomData.Tile2ScrY(Cur->y);
      if (x>-Tile_W && x<320 && y>-Tile_H && y<200)
      {
        if (x>0 && x<320-Tile_W && y>0 && y<200-Tile_H)
          g_tBitmap(x, y, &Tiles[Cur->frame]); else
          g_ctBitmap(x, y, &Tiles[Cur->frame]);
      } // don't draw way out of bounds
      Cur = Cur->next;
    }
  }

  if (RoomCur->RoomCeiling && kDrawCeiling) RoomData.DrawRoof();

  // display pending barks, messages, etc... (but don't update them)
  Cur = RoomCur->ObjHead;
  g_Font(2);
  while (Cur)
  {
    if (Cur->barktime)
    {
      g_FontColor(Cur->barktime % 16);
      int bx = RoomData.Tile2ScrX(Cur->x);
      int by = RoomData.Tile2ScrY(Cur->y);
      if (by<8)
        g_sprintf(bx+Tile_W+2, by, Cur->bark); else
        g_sprintf(bx, by-8, Cur->bark);
    }
    Cur = Cur->next;
  }

  if (msgtime)
  {
    g_Font(1);
    g_FontColor(msgtime % 16);
    switch (barktype)
    {
      case 2 : // message
        g_scenprintf(190, msg);
        break;
      case 3 : // mono-spaced message
        g_mscenprintf(190, msg);
        break;
    }
  }
}

void ResetCamera(void)
{
  // focus camera on first player
  Player = FindObj("Player", RoomCur->ObjHead);
  if (Player)
  {
    CamFocus = Player;
    CamX = Player->x;
    CamY = Player->y;
  } else
  {
    DebugBark("No player in room!");
    CamFocus = NULL;
    CamX = 0;
    CamY = 0;
  }
}

void KernelSwitchRoom(pRoom r)
{
  if (r)
  {
    RoomCur = r;
    RoomData.InitMap(r->RoomW, r->RoomH, r->RoomFloor, r->RoomObject, r->RoomCeiling);
  }

  ResetCamera();

  if (!r) return;

  if (stricmp(r->Mod, "None")==0) return;
  if (stricmp(r->Mod, "Silence")==0)
  {
    zm_modStop();
    return;
  }

  switch (zm_modLoad(r->Mod))
  {
    case 0 :
      zm_modStart();
      break;
    case -1 : case -2 :
      DebugBark("Error loading room music");
      zm_modStop();
      break;
  }
}

void Bark(int t, pObject o, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(buf, fmt, arglist);
  va_end(arglist);

  strncpy(o->bark, buf, BARK_MAX);
  o->barktime = t;
}

void Message(int k, int t, char *fmt, ...)
{
  va_list arglist;

  va_start(arglist, fmt);
  vsprintf(buf, fmt, arglist);
  va_end(arglist);

  barktype = k;

  switch (barktype)
  {
    case BK_MSG :
    case BK_MMSG :
      strncpy(msg, buf, 512);
      msgtime = t;
      break;
  }
}

void DebugBark(char *fmt, ...)
{
  if (debug)
  {
    va_list arglist;

    va_start(arglist, fmt);
    vsprintf(buf, fmt, arglist);
    va_end(arglist);

    barktype = BK_DEF;
    strncpy(msg, buf, 512);
    msgtime = 90;
  }
}

int Move(int dx, int dy, pObject o, pRoom r=RoomCur)
{
  if (!o) return BAD_OBJ;
  
  if (r->RoomObject[dy*r->RoomW+dx]) return 0;
  pObject obstruct = GetObjXY(dx, dy, r);
  if (!obstruct->objPlayer && obstruct) return 0;

  if (r != RoomCur && o==Player)
  {
    KernelSwitchRoom(r);
    Player->x = dx;
    Player->y = dy;
    CamFocus = Player;
    CamX = Player->x;
    CamY = Player->y;
  } else
  {
    o->x = dx;
    o->y = dy;
  }

  return 1;
}

// Go(int d, pObject o)
//
// Handles all cases of all objects interacting with each other
// Tested:
//
// Returns:
//  BAD_OBJ - Bad object passed
//  0       - Go blocked
//  1       - Go successful

int Go(int d, pObject o)
{
  if (!o) return BAD_OBJ;         // if invalid object, don't bother

  // First, we set some variables up to use throughout the function.
  // x, y is the object's original position.
  // dx, dy will be the object's final position.
  int x = o->x;
  int y = o->y;
  int dx = x;
  int dy = y;

  // Determine object's final position, and save coordinates to dx, dy
  switch (d)
  {
    case GO_NORTH : dy--; break;
    case GO_SOUTH : dy++; break;
    case GO_EAST  : dx++; break;
    case GO_WEST  : dx--; break;
  }

  // Set up some quick-access width/height variables
  int w = RoomCur->RoomW;
  int h = RoomCur->RoomH;

  // Check THUD between object and room edges. If the player hits a room
  // edge, we will switch rooms. If an object hits a room edge, it will
  // signal a THUD.
  // FIXME: Allow objects to traverse room boundaries with player?

  if (dy<0 || dy>=h || dx<0 || dx>=w)
  {
    if (o==Player)
    {
      if (dy<0 && RoomCur->LinkN)       // north exit
      {
        Move(x, RoomCur->LinkN->RoomH-1, o, RoomCur->LinkN);
        return 1;
      } else
      if (dy>=h && RoomCur->LinkS)      // south exit
      {
        Move(x, 0, o, RoomCur->LinkS);
        return 1;
      } else
      if (dx<0 && RoomCur->LinkW)       // west exit
      {
        Move(RoomCur->LinkW->RoomW-1, y, o, RoomCur->LinkW);
        return 1;
      } else
      if (dx>=w && RoomCur->LinkE)      // east exit
      {
        Move(0, y, o, RoomCur->LinkE);
        return 1;
      }
    }
    SendMessage(o, "thud");
    DebugBark("%s thuds a room edge", o->name);
    return 2;
  }

  // Check THUD with walls. If any object hits a wall, it will be sent a
  // THUD signal.

  short *objl = RoomCur->RoomObject;
  if (objl[dy*w+dx])
  {
    SendMessage(o, "thud");
    DebugBark("%s thuds a wall", o->name);
//    return 0;
    return 2;
  }

  // Check THUD with other objects. If the player hits a target object, the
  // target object will receive a TOUCH message. If a source object hits a
  // (target) object then the source object will receive a THUD message.
  // Pushing rules will go into effect.

  pObject touched = GetObjXY(dx, dy, RoomCur);
  if (touched)
  {
    if (o==Player)
    {
      SendMessage(touched, "touch");
      if (touched->objInvisible) touched->objInvisible = 0;
      DebugBark("%s touches %s", o->name, touched->name);
    }

    int allowpush=1;
    if (o==Player && !touched->objPushP) allowpush = 0;
    if (!o->objPusher) allowpush = 0;
    if (allowpush)
    {
      int vt = d==GO_NORTH || d==GO_SOUTH;
      int hz = d==GO_EAST || d==GO_WEST;
      if ( (touched->objPushNS && vt) ||
           (touched->objPushEW && hz) )
      {
        if (Go(d, touched) == 1)
        {
          o->x = dx;
          o->y = dy;
//        DebugBark("Let's kick the tires, and light the fires!");
          return 1;
        }
      }
    }
    SendMessage(o, "thud");
    DebugBark("%s thuds %s", o->name, touched->name);
    return 2;
  }

  o->x = dx;
  o->y = dy;
  return 1;
}

int ProcessOkCanKernel(char *str)
{
  Rect r;
  SetupWindowKernel();

  int len = g_strlen(str)+10;
  if (len<135) len = 135;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-125)/2), Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  Win.Done();
  UpdateGame();
  mk_Clear();

  return (Result==1);
}

void ProcessDebug(void)
{
  pObject o = RoomCur->ObjHead;
  Rect r;
  SetupWindowKernel();

  r.Set(200, 95);
  Dialog Win(r);
  Win.SetTitle("Debug Window");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-30, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Set(Win.r.x+5, Win.r.y2-15, 40, 12);
  Button Room(r);
  Room.SetTitle("Room");

  r.MoveRel(45, 0);
  Button Glob(r);
  Glob.SetTitle("Global");

  r.MoveRel(45, 0);
  Button Syst(r);
  Syst.SetTitle("System");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-65, Win.r.y+35);
  Input OName(r);
  OName.SetTitle("Object name:");
  if (o)
    OName.SetData(o->name); else
    OName.SetData("No objects in this room.");
  OName.SetMax(RNAME_MAX);
  OName.AllowEdit(0);

  r.MoveRel(OName.r.w+5, -2);
  r.Size(55, 12);
  Button OSelect(r);
  OSelect.SetTitle("Select...");

  Win.Add(&OName);
  Win.Add(&OSelect);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&Room);
  Win.Add(&Glob);
  Win.Add(&Syst);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {}

  Win.Done();
  UpdateGame();
  mk_Clear();
}

short Play(pRoom r)
{
  int st, et, error;
  char k[] = "key_???";
  char c;

  KernelSwitchRoom(r);
  
  UpdateGame();
  zzt_Fade(Black, zztPal);

  quitplaying = 0;

  kDrawFloor = kDrawObject = kDrawCeiling = 1;

  do {
    // send certain keypresses to all global objects
    c = mk_PollChar();
    if (c != mcNone)
    {
      switch (c)
      {
        case mcF1 :
          if (mk_Shift()) SendGlobalMsg("key_sf1"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf1"); else
          if (mk_Alt()) SendGlobalMsg("key_af1"); else
            SendGlobalMsg("key_f1");
          break;
        case mcF2 :
          if (mk_Shift()) SendGlobalMsg("key_sf2"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf2"); else
          if (mk_Alt()) SendGlobalMsg("key_af2"); else
            SendGlobalMsg("key_f2");
          break;
        case mcF3 :
          if (mk_Shift()) SendGlobalMsg("key_sf3"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf3"); else
          if (mk_Alt()) SendGlobalMsg("key_af3"); else
            SendGlobalMsg("key_f3");
          break;
        case mcF4 :
          if (mk_Shift()) SendGlobalMsg("key_sf4"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf4"); else
          if (mk_Alt()) SendGlobalMsg("key_af4"); else
            SendGlobalMsg("key_f4");
          break;
        case mcF5 :
          if (mk_Shift()) SendGlobalMsg("key_sf5"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf5"); else
          if (mk_Alt()) SendGlobalMsg("key_af5"); else
            SendGlobalMsg("key_f5");
          break;
        case mcF6 :
          if (mk_Shift()) SendGlobalMsg("key_sf6"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf6"); else
          if (mk_Alt()) SendGlobalMsg("key_af6"); else
            SendGlobalMsg("key_f6");
          break;
        case mcF7 :
          if (mk_Shift()) SendGlobalMsg("key_sf7"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf7"); else
          if (mk_Alt()) SendGlobalMsg("key_af7"); else
            SendGlobalMsg("key_f7");
          break;
        case mcF8 :
          if (mk_Shift()) SendGlobalMsg("key_sf8"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf8"); else
          if (mk_Alt()) SendGlobalMsg("key_af8"); else
            SendGlobalMsg("key_f8");
          break;
        case mcF9 :
          if (mk_Shift()) SendGlobalMsg("key_sf9"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf9"); else
          if (mk_Alt()) SendGlobalMsg("key_af9"); else
            SendGlobalMsg("key_f9");
          break;
        case mcF10 :
          if (mk_Shift()) SendGlobalMsg("key_sf10"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf10"); else
          if (mk_Alt()) SendGlobalMsg("key_af10"); else
            SendGlobalMsg("key_f10");
          break;
        case mcF11 :
          if (mk_Shift()) SendGlobalMsg("key_sf11"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf11"); else
          if (mk_Alt()) SendGlobalMsg("key_af11"); else
            SendGlobalMsg("key_f11");
          break;
        case mcF12 :
          if (mk_Shift()) SendGlobalMsg("key_sf12"); else
          if (mk_Ctrl()) SendGlobalMsg("key_cf12"); else
          if (mk_Alt()) SendGlobalMsg("key_af12"); else
            SendGlobalMsg("key_f12");
          break;
        case mcEnter :
          SendGlobalMsg("key_enter");
          break;
        default :
          if (mk_Ctrl())
          {
            k[4] = 'c';
            k[5] = c;
          } else
          if (mk_Alt())
          {
            k[4] = 'a';
            k[5] = c;
          } else k[4] = c;
          SendGlobalMsg(k);
          break;
      }
    }

    if (mk_Shift())
    {
      if (mk_Down(mkUp)) SendPlayerMsg("shoot_n");
      if (mk_Down(mkDown)) SendPlayerMsg("shoot_s");
      if (mk_Down(mkLeft)) SendPlayerMsg("shoot_w");
      if (mk_Down(mkRight)) SendPlayerMsg("shoot_e");
    } else
    {
      if (mk_Down(mkUp)) SendPlayerMsg("go_n");
      if (mk_Down(mkDown)) SendPlayerMsg("go_s");
      if (mk_Down(mkLeft)) SendPlayerMsg("go_w");
      if (mk_Down(mkRight)) SendPlayerMsg("go_e");
    }

    st = clock();
    Update();
    UpdateCam();
    UpdateGame();

    if (!hyper)
    {
      do {
        et = clock();
      } while (et-st<1);
    }

    if (mk_Alt(mkZ))
    {
      mk_Clear();
      ProcessDebug();
    }

    if (mk_Pressed(mkZ))
    {
      mk_Clear();
      hyper = !hyper;
    }

    if (mk_Pressed(mkX))
    {
      mk_Clear();
      debug = !debug;
    }

    if (mk_Pressed(mkEsc))
    {
      if (ineditor)
        quitplaying = 1; else
        quitplaying = ProcessOkCanKernel("I knew it. A quitter.");
    }
  } while (!quitplaying);

  zzt_Fade(zztPal, Black);

  return RoomCur->NameHash;
}

