// Mem32, version 1.0 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version, highly optimized 32-bit memset variations
//                Please read tech notes in this file for more information
// Propose 1.1 -- Add memcpy substitute

#ifndef __mem32__
#define __mem32__

#include <stdlib.h>

// This is *solid* 32-bit inline code which is more than 100% faster than
// Watcom's memset routine. But, you must first know how and when to use
// each individual type.

// memset8 can safely be substituted for memset, it works exactly like it,
// only twice as fast. As you port your software to use this code, first try
// substituting all memset's with memset8's. Then examine how each of the
// other routines work and apply them appropriately for even better speed.
extern void memset8(void *s, unsigned char c, size_t n);
#pragma aux memset8 = \
  "mov ebx, eax"\
  "shl ebx, 8"\
  "or  eax, ebx"\
  "mov ebx, eax"\
  "shl ebx, 16"\
  "or  eax, ebx"\
  "shr ecx, 1"\
  "jnc next1"\
  "stosb"\
  "next1:"\
  "shr ecx, 1"\
  "jnc next2"\
  "stosw"\
  "next2:"\
  "rep stosd"\
parm [edi] [eax] [ecx]\
modify [ebx];

// memcpy8 can safely be substituted for memcpy, it works exactly like it,
// only twice as fast. As you port your software to use this code, first try
// substituting all memcpy's with memcpy8's.
extern void memcpy8(void *dst, const void *src, size_t n);
#pragma aux memcpy8 = \
  "shr ecx, 1"\
  "jnc next1"\
  "movsb"\
  "next1:"\
  "shr ecx, 1"\
  "jnc next2"\
  "movsw"\
  "next2:"\
  "rep movsd"\
parm [edi] [esi] [ecx];

// memcpy32 copies like memcpy, but has a granularity of 4, and only copies
// ranges divisible by 4.
extern void memcpy32(void *dst, const void *src, size_t n);
#pragma aux memcpy32 = \
  "rep movsd"\
parm [edi] [esi] [ecx];

// memset16 repeats an unsigned short (something memset can't do.) I've
// used it in 64k color graphics software for near 8-bit graphics speed. Also
// excellent for software utilizing the Unicode string tools (16-bit chars)
// granularity of 2!
extern void memset16(void *s, unsigned short c, size_t n);
#pragma aux memset16 = \
  "mov ebx, eax"\
  "shl ebx, 16"\
  "or  eax, ebx"\
  "shr ecx, 2"\
  "jnc next"\
  "stosw"\
  "next:"\
  "rep stosd"\
parm [edi] [eax] [ecx]\
modify [ebx];

// memset32 repeats an unsigned long (something memset can't do.) I've
// used it in 4G color graphics software for near 8-bit graphics speed.
// granularity of 4!
extern void memset32(void *s, unsigned long c, size_t n);
#pragma aux memset32 = \
  "shr ecx, 2"\
  "rep stosd"\
parm [edi] [eax] [ecx];

// clear8 is excellent for 8-bit (256 color) graphics software which require
// a screen clear per frame. This requires the inputted size to already be
// divided by 4. It also has a minimum granularity of 4, which makes it
// difficult to use with strings, but excellent for graphics screens (which
// to my knowledge always have a granularity of 4.)
extern void clear8(void *s, unsigned char c, size_t n);  // n must be div by 4
#pragma aux clear8 = \
  "mov ebx, eax"\
  "shl ebx, 8"\
  "or  eax, ebx"\
  "mov ebx, eax"\
  "shl ebx, 16"\
  "or  eax, ebx"\
  "rep stosd"\
parm [edi] [eax] [ecx]\
modify [ebx];

// clear16 is excellent for 16-bit (64k color) graphics software which require
// a screen clear per frame. This requires the inputted size to already be
// divided by 4. It also has a minimum granularity of 4, which makes it
// difficult to use with strings, but excellent for graphics screens (which
// to my knowledge always have a granularity of 4.)
extern void clear16(void *s, unsigned short c, size_t n);  // n must be div by 4
#pragma aux clear16 = \
  "mov ebx, eax"\
  "shl ebx, 16"\
  "or  eax, ebx"\
  "shr ecx, 2"\
  "rep stosd"\
parm [edi] [eax] [ecx]\
modify [ebx];

// clear32 is excellent for 32-bit (4G color) graphics software which require
// a screen clear per frame. This requires the inputted size to already be
// divided by 4. It also has a minimum granularity of 4, which makes it
// difficult to use with strings, but excellent for graphics screens (which
// to my knowledge always have a granularity of 4.)
extern void clear32(void *s, unsigned long c, size_t n);  // n must be div by 4
#pragma aux clear32 = \
  "shr ecx, 2"\
  "rep stosd"\
parm [edi] [eax] [ecx];

// blackout is excellent for QUICKLY blacking out a graphics screen (once
// again, a minimum granularity of 4)
extern void blackout(void *s, size_t n);  // n must be div by 4
#pragma aux blackout = \
  "xor eax, eax"\
  "rep stosd"\
parm [edi] [ecx]\
modify [eax];

#endif
