// Palette Manager, version 2.01 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 2.0 -- Rewritten with more powerful palette management functions
// Version 2.01 - Added "hue" feature to GetIndex (for percentages over
//                100%, colors begin hueing)

// Propose 2.1 -- Add ability to load various different formats of palettes,
//                and embedded palettes including Windows, JASC, PCX, GIF,
//                etc...
// Propose 2.5 -- A virtual palette to allow screen fades, rotations, etc.
//                in 16-bit color modes. (This will be limited to linear
//                modes, because it requires going through framebuffer and
//                physically changing values. It may be too slow to be
//                worthwhile, but will look into it.)

#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <conio.h>
#include <i86.h>
#include "mem32.h"
#include "pal.h"
#include "vgl.h"

PALETTE VGA, Cur, Black;
INDEX   Dark, Lite;
PALETTE temp;
int Speed;

void pal_Init(void)
{
  for (int i=0; i<256; i++)
  {
    outp(0x3C7, i);
    VGA[i][0] = inp(0x3C9);
    VGA[i][1] = inp(0x3C9);
    VGA[i][2] = inp(0x3C9);
  }
  memcpy8(Cur, VGA, sizeof(PALETTE));
  memset8(Black, 0, sizeof(PALETTE));
  memset8(temp, 0, sizeof(PALETTE));
  pal_GetIndex(60, 60, 60, Dark);
  pal_GetIndex(120, 120, 120, Lite);
  pal_SetSpeed(100);
}

void pal_SetSpeed(int s)
{
  Speed = s;
}

int pal_GetSpeed(void)
{
  return Speed;
}

void pal_SetRGB(int c, char r, char g, char b)
{
  outp(0x3C8, c);
  outp(0x3C9, r);
  outp(0x3C9, g);
  outp(0x3C9, b);
}

void pal_GetRGB(int c, char *r, char *g, char *b)
{
  outp(0x3C7, c);
  *r = inp(0x3C9);
  *g = inp(0x3C9);
  *b = inp(0x3C9);
}

void pal_Load(char *f, PALETTE p)
{
  ifstream fp;
  int s;

  fp.open(f, ios::in | ios::binary);
  if (!fp)
  {
    g_Done();
    printf("Palette Manager: Error opening palette file %s.\n", f);
    exit(0);
  }

  fp.read((unsigned char *)p, 768);
  s = fp.gcount();
  if (s != 768)
  {
    g_Done();
    printf("Palette Manager: Error reading palette file %s (read %d instead of 768).\n", f, s);
    exit(0);
  }

  fp.close();
}

void pal_Save(char *f, PALETTE p)
{
  ofstream fp;
  int s;

  fp.open(f, ios::out | ios::binary);
  if (!fp)
  {
    g_Done();
    printf("Palette Manager: Error writing palette file %s.\n", f);
    exit(0);
  }

  fp.write((unsigned char *)p, 768);
  fp.close();
}

void pal_SetCur(PALETTE p)
{
  memcpy8(Cur, p, sizeof(PALETTE));
  pal_Set(Cur);
  pal_GetIndex(60, 60, 60, Dark);
  pal_GetIndex(120, 120, 120, Lite);
}

void pal_Set(PALETTE p, int f, int l)
{
  g_VSync();
  for (int i=f; i<=l; i++)
    pal_SetRGB(i, p[i][0], p[i][1], p[i][2]);
}

void pal_Get(PALETTE *p)
{
  for (int i=0; i<256; i++)
    pal_GetRGB(i, p[i][0], p[i][1], p[i][2]);
}

void pal_SetIntensity(PALETTE sp, PALETTE dp, int i, int f, int l)
{
  for (int z=f; z<=l; z++)
  {
    temp[z][0] = ((sp[z][0]*(Speed-i)) + (dp[z][0] * i)) / Speed;
    temp[z][1] = ((sp[z][1]*(Speed-i)) + (dp[z][1] * i)) / Speed;
    temp[z][2] = ((sp[z][2]*(Speed-i)) + (dp[z][2] * i)) / Speed;
  }
}

#define RED_B 0.299
#define GRN_B 0.587
#define BLU_B 0.114

char pal_FindColor(int r1, int g1, int b1)
{
  float c1, c2, d0, d1, d2, d3;
  int high, test, r2, g2, b2;
  char index = 0;

  high = INT_MAX;

  for (int i=0; i<256; i++)
  {
    r2 = Cur[i][0];
    g2 = Cur[i][1];
    b2 = Cur[i][2];

    c1 = (RED_B * r1) + (GRN_B * g1) + (BLU_B * b1);
    c2 = (RED_B * r2) + (GRN_B * g2) + (BLU_B * b2);

    d0 = (r1-r2) * (r1-r2);
    d1 = (g1-g2) * (g1-g2);
    d2 = (b1-b2) * (b1-b2);
    d3 = (c1-c2) * (c1-c2);

    test = d3 + d0 + d1 + d2;

    if (test<high)
    {
      high  = test;
      index = i;
    }
  }

  return index;
}

void pal_GetIndex(int r, int g, int b, INDEX p)
{
  int ar, ag, ab, ra, ga, ba;

  if (r>100) ra = r-100; else ra = 0;
  if (g>100) ga = g-100; else ga = 0;
  if (b>100) ba = b-100; else ba = 0;

  for (int i=0; i<256; i++)
  {
    ar = ((Cur[i][0]*r)/100)+ra;
    ag = ((Cur[i][1]*g)/100)+ga;
    ab = ((Cur[i][2]*b)/100)+ba;

    if (ar<0) ar = 0;
    if (ag<0) ag = 0;
    if (ab<0) ab = 0;
    if (ar>63) ar = 63;
    if (ag>63) ag = 63;
    if (ab>63) ab = 63;

    p[i] = pal_FindColor(ar, ag, ab);
  }
}

void pal_CrossFade(PALETTE sp, PALETTE dp, int f, int l)
{
  for (int i=0; i<=Speed; i++)
  {
    pal_SetIntensity(sp, dp, i, f, l);
    pal_Set(temp);
  }
}

void pal_Rotate(int dir, int num, int f, int l)
{
  char triad[3];
  int i;

  switch (dir)
  {
    case FORWARD :
      for (i=0; i<num; i++)
      {
        memcpy8(triad, Cur[l], sizeof(triad));
        memcpy8(temp[f+1], Cur, (l-f)*sizeof(triad));
        memcpy8(temp, triad, sizeof(triad));
        memcpy8(Cur, temp, sizeof(PALETTE));
        pal_Set(Cur);
      }
      break;
    case BACKWARD :
      for (i=0; i<num; i++)
      {
        memcpy8(triad, Cur[f], sizeof(triad));
        memcpy8(temp, Cur[f+1], (l-f)*sizeof(triad));
        memcpy8(temp[l], triad, sizeof(triad));
        memcpy8(Cur, temp, sizeof(PALETTE));
        pal_Set(Cur);
      }
      break;
  }
  pal_SetCur(temp);
}
